//#####################################################################
// Copyright 2002-2006, Robert Bridson, Ronald Fedkiw, Eran Guendelman, Geoffrey Irving, Neil Molino, Avi Robinson-Mosher, Andrew Selle, Joseph Teran.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class TETRAHEDRON
//##################################################################### 
#include <PhysBAM_Tools/Arrays/ARRAY.h>
#include <PhysBAM_Tools/Arrays/ARRAY_VIEW.h>
#include <PhysBAM_Tools/Arrays/INDIRECT_ARRAY.h>
#include <PhysBAM_Tools/Grids_Uniform/GRID.h>
#include <PhysBAM_Tools/Interpolation/LINEAR_INTERPOLATION.h>
#include <PhysBAM_Tools/Vectors/VECTOR_3D.h>
#include <PhysBAM_Geometry/Basic_Geometry/TETRAHEDRON.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_3D.h>
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class T> TETRAHEDRON<T>::
TETRAHEDRON()
    :X(TV(),TV(0,1,0),TV(1,0,0),TV(0,0,-1))
{
    Create_Triangles();
}
//#####################################################################
// Constructor
//#####################################################################
template<class T> TETRAHEDRON<T>::
TETRAHEDRON(const TV& x1_input,const TV& x2_input,const TV& x3_input,const TV& x4_input)
    :X(x1_input,x2_input,x3_input,x4_input)
{
    Create_Triangles();
}
//#####################################################################
// Function Create_Triangles
//#####################################################################
template<class T> void TETRAHEDRON<T>::
Create_Triangles()
{
    if(TV::Dot_Product(TV::Cross_Product(X[2]-X[1],X[3]-X[1]),X[4]-X[1]) <= 0){
        triangle1.Specify_Three_Points(X[1],X[2],X[3]);triangle2.Specify_Three_Points(X[1],X[4],X[2]);
        triangle3.Specify_Three_Points(X[1],X[3],X[4]);triangle4.Specify_Three_Points(X[2],X[4],X[3]);}
    else{
        triangle1.Specify_Three_Points(X[1],X[3],X[2]);triangle2.Specify_Three_Points(X[1],X[2],X[4]);
        triangle3.Specify_Three_Points(X[1],X[4],X[3]);triangle4.Specify_Three_Points(X[2],X[3],X[4]);}
}
//#####################################################################
// Function Normal
//#####################################################################
template<class T> VECTOR<T,3> TETRAHEDRON<T>::
Normal(const VECTOR<T,3>& location,const int aggregate) const
{
    assert(aggregate >= 1 && aggregate <= 4);
    if(aggregate == 1) return triangle1.normal;
    else if(aggregate == 2) return triangle2.normal;
    else if(aggregate == 3) return triangle3.normal;
    else return triangle4.normal;
}
//#####################################################################
// Function Inside
//#####################################################################
template<class T> bool TETRAHEDRON<T>::
Inside(const VECTOR<T,3>& location,const T thickness_over_two) const
{
    if(triangle1.Inside(location,thickness_over_two) && triangle2.Inside(location,thickness_over_two) && triangle3.Inside(location,thickness_over_two) && triangle4.Inside(location,thickness_over_two)) 
        return true;
    else return false;
}
//#####################################################################
// Function Outside
//#####################################################################
template<class T> bool TETRAHEDRON<T>::
Outside(const VECTOR<T,3>& location,const T thickness_over_two) const
{
    return triangle1.Outside(location,thickness_over_two) || triangle2.Outside(location,thickness_over_two) || triangle3.Outside(location,thickness_over_two)
        || triangle4.Outside(location,thickness_over_two);
}
//#####################################################################
// Function Boundary
//#####################################################################
template<class T> bool TETRAHEDRON<T>::
Boundary(const VECTOR<T,3>& location,const T thickness_over_two) const
{
    return !Inside(location,thickness_over_two) && !Outside(location,thickness_over_two);
}
//#####################################################################
// Function Thickened
//#####################################################################
template<class T> TETRAHEDRON<T> TETRAHEDRON<T>::
Thickened(const T thickness_over_two) const
{
    assert(Signed_Volume()>0);
    T m1=thickness_over_two/triangle4.Signed_Distance(X[1]),m2=thickness_over_two/triangle3.Signed_Distance(X[2]),m3=thickness_over_two/triangle2.Signed_Distance(X[3]),m4=thickness_over_two/triangle1.Signed_Distance(X[4]);
    return TETRAHEDRON<T>(Point_From_Barycentric_Coordinates(VECTOR<T,3>(m1,m2,m3)),Point_From_Barycentric_Coordinates(VECTOR<T,3>((T)1-m2-m3-(T)root_three*m4,m2,m3)),
                          Point_From_Barycentric_Coordinates(VECTOR<T,3>(m1,(T)1-m1-m3-(T)root_three*m4,m3)),Point_From_Barycentric_Coordinates(VECTOR<T,3>(m1,m2,(T)1-m1-m2-(T)root_three*m4)));
}
//#####################################################################
// Function Bounding_Box
//#####################################################################
template<class T> RANGE<VECTOR<T,3> > TETRAHEDRON<T>::
Bounding_Box() const
{      
    return RANGE<TV>::Bounding_Box(X);
}
//#####################################################################
// Function Surface
//#####################################################################
template<class T> VECTOR<T,3> TETRAHEDRON<T>::
Surface(const VECTOR<T,3>& location) const
{      
    if(Inside(location)){
        int triangle=1;T distance=VECTOR<T,3>::Dot_Product(triangle1.x1-location,triangle1.normal);
        T distance_temp=VECTOR<T,3>::Dot_Product(triangle2.x1-location,triangle2.normal);
        if(distance_temp < distance){triangle=2;distance=distance_temp;}
        distance_temp=VECTOR<T,3>::Dot_Product(triangle3.x1-location,triangle3.normal);
        if(distance_temp < distance){triangle=3;distance=distance_temp;}
        distance_temp=VECTOR<T,3>::Dot_Product(triangle4.x1-location,triangle4.normal);
        if(distance_temp < distance){triangle=4;distance=distance_temp;}
        switch(triangle){
            case 1:return location+distance*triangle1.normal;
            case 2:return location+distance*triangle2.normal;
            case 3:return location+distance*triangle3.normal;
            case 4:return location+distance*triangle4.normal;
            default:return VECTOR<T,3>(0,0,0);}} // should never be called!
    else{
        VECTOR<T,3> surface_point(location);
        if(triangle1.Lazy_Outside(surface_point)) surface_point-=VECTOR<T,3>::Dot_Product(location-triangle1.x1,triangle1.normal)*triangle1.normal;
        if(triangle2.Lazy_Outside(surface_point)) surface_point-=VECTOR<T,3>::Dot_Product(location-triangle2.x1,triangle2.normal)*triangle2.normal;
        if(triangle3.Lazy_Outside(surface_point)) surface_point-=VECTOR<T,3>::Dot_Product(location-triangle3.x1,triangle3.normal)*triangle3.normal;
        if(triangle4.Lazy_Outside(surface_point)) surface_point-=VECTOR<T,3>::Dot_Product(location-triangle4.x1,triangle4.normal)*triangle4.normal;
        return surface_point;}
}
//#####################################################################
// Function Closest_Point
//#####################################################################
template<class T> VECTOR<T,3> TETRAHEDRON<T>::
Closest_Point(const VECTOR<T,3>& location,VECTOR<T,3>& weights) const
{      
    if(Inside(location)){weights=First_Three_Barycentric_Coordinates(location);return location;}
    
    VECTOR<T,3> triangle_weights,triangle_closest_point=TRIANGLE_3D<T>(X[1],X[2],X[3]).Closest_Point(location,triangle_weights),closest_point=triangle_closest_point;
    weights=VECTOR<T,3>(triangle_weights.x,triangle_weights.y,triangle_weights.z);
    T triangle_distance_squared=(closest_point-location).Magnitude_Squared(),distance_squared=triangle_distance_squared;
    
    triangle_closest_point=TRIANGLE_3D<T>(X[1],X[2],X[4]).Closest_Point(location,triangle_weights);triangle_distance_squared=(triangle_closest_point-location).Magnitude_Squared();
    if(triangle_distance_squared<distance_squared){
        weights=VECTOR<T,3>(triangle_weights.x,triangle_weights.y,0);closest_point=triangle_closest_point;distance_squared=triangle_distance_squared;}
    
    triangle_closest_point=TRIANGLE_3D<T>(X[1],X[3],X[4]).Closest_Point(location,triangle_weights);triangle_distance_squared=(triangle_closest_point-location).Magnitude_Squared();
    if(triangle_distance_squared<distance_squared){
        weights=VECTOR<T,3>(triangle_weights.x,0,triangle_weights.y);closest_point=triangle_closest_point;distance_squared=triangle_distance_squared;}
    
    triangle_closest_point=TRIANGLE_3D<T>(X[2],X[3],X[4]).Closest_Point(location,triangle_weights);triangle_distance_squared=(triangle_closest_point-location).Magnitude_Squared();
    if(triangle_distance_squared<distance_squared){
        weights=VECTOR<T,3>(0,triangle_weights.x,triangle_weights.y);return triangle_closest_point;}
        
    return closest_point;
}
//#####################################################################
// Function Volume
//#####################################################################
template<class T> T TETRAHEDRON<T>::
Volume() const
{
    return (T)one_sixth*abs(VECTOR<T,3>::Dot_Product(VECTOR<T,3>::Cross_Product(X[2]-X[1],X[3]-X[1]),X[4]-X[1]));
}
//#####################################################################
// Function Signed_Volume
//#####################################################################
template<class T> T TETRAHEDRON<T>::
Signed_Volume() const
{  
    return (T)one_sixth*VECTOR<T,3>::Dot_Product(VECTOR<T,3>::Cross_Product(X[2]-X[1],X[3]-X[1]),X[4]-X[1]); 
}
//#####################################################################
// Function Minimum_Angle
//#####################################################################
template<class T> T TETRAHEDRON<T>::
Minimum_Angle() const
{  
    return min(triangle1.Minimum_Angle(),triangle2.Minimum_Angle(),triangle3.Minimum_Angle(),triangle4.Minimum_Angle());
}
//#####################################################################
// Function Maximum_Angle
//#####################################################################
template<class T> T TETRAHEDRON<T>::
Maximum_Angle() const
{  
    return max(triangle1.Maximum_Angle(),triangle2.Maximum_Angle(),triangle3.Maximum_Angle(),triangle4.Maximum_Angle());
}
//#####################################################################
// Function Minimum_Altitude
//#####################################################################
template<class T> T TETRAHEDRON<T>::
Minimum_Altitude() const
{  
    return min(abs(triangle1.Signed_Distance(X[4])),abs(triangle2.Signed_Distance(X[3])),abs(triangle3.Signed_Distance(X[2])),abs(triangle4.Signed_Distance(X[1])));
}
//#####################################################################
// Function Aspect_Ratio
//#####################################################################
template<class T> T TETRAHEDRON<T>::
Aspect_Ratio() const
{  
    return Maximum_Edge_Length()/Minimum_Altitude();
}
//#####################################################################
// Function Minimum_Dihedral_Angle
//#####################################################################
template<class T> T TETRAHEDRON<T>::
Minimum_Dihedral_Angle() const
{  
    return acos(-min(VECTOR<T,3>::Dot_Product(triangle1.normal,triangle2.normal),VECTOR<T,3>::Dot_Product(triangle1.normal,triangle3.normal),
                               VECTOR<T,3>::Dot_Product(triangle1.normal,triangle4.normal),VECTOR<T,3>::Dot_Product(triangle2.normal,triangle3.normal),
                               VECTOR<T,3>::Dot_Product(triangle2.normal,triangle4.normal),VECTOR<T,3>::Dot_Product(triangle3.normal,triangle4.normal))); 
}
//#####################################################################
// Function Maximum_Dihedral_Angle
//#####################################################################
template<class T> T TETRAHEDRON<T>::
Maximum_Dihedral_Angle() const
{  
    return acos(-max(VECTOR<T,3>::Dot_Product(triangle1.normal,triangle2.normal),VECTOR<T,3>::Dot_Product(triangle1.normal,triangle3.normal),
                               VECTOR<T,3>::Dot_Product(triangle1.normal,triangle4.normal),VECTOR<T,3>::Dot_Product(triangle2.normal,triangle3.normal),
                               VECTOR<T,3>::Dot_Product(triangle2.normal,triangle4.normal),VECTOR<T,3>::Dot_Product(triangle3.normal,triangle4.normal)));
}
//#####################################################################
// Function Signed_Reciprocal_Aspect_Ratio
//#####################################################################
template<class T> T TETRAHEDRON<T>::
Signed_Reciprocal_Aspect_Ratio(const VECTOR<T,3>& x1,const VECTOR<T,3>& x2,const VECTOR<T,3>& x3,const VECTOR<T,3>& x4)
{
    return min(VECTOR<T,3>::Dot_Product(x4-x1,PLANE<T>::Normal(x1,x2,x3)),VECTOR<T,3>::Dot_Product(x3-x1,PLANE<T>::Normal(x1,x4,x2)),
                     VECTOR<T,3>::Dot_Product(x2-x1,PLANE<T>::Normal(x1,x3,x4)),VECTOR<T,3>::Dot_Product(x1-x2,PLANE<T>::Normal(x2,x4,x3)))/
              max((x1-x2).Magnitude(),(x1-x3).Magnitude(),(x1-x4).Magnitude(),(x2-x3).Magnitude(),(x2-x4).Magnitude(),(x3-x4).Magnitude());
}
//#####################################################################
template class TETRAHEDRON<float>;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class TETRAHEDRON<double>;
#else
template double TETRAHEDRON<double>::Minimum_Altitude() const; // used by MATRIX<double,3>
#endif
